"use strict";(self.webpackChunknbdime_webapp=self.webpackChunknbdime_webapp||[]).push([[3405],{83405:(r,n,t)=>{function e(r,n){return r.skipToEnd(),n.cur=p,"error"}function u(r,n){return r.match(/^HTTP\/\d\.\d/)?(n.cur=c,"keyword"):r.match(/^[A-Z]+/)&&/[ \t]/.test(r.peek())?(n.cur=i,"keyword"):e(r,n)}function c(r,n){var t=r.match(/^\d+/);if(!t)return e(r,n);n.cur=o;var u=Number(t[0]);return u>=100&&u<400?"atom":"error"}function o(r,n){return r.skipToEnd(),n.cur=p,null}function i(r,n){return r.eatWhile(/\S/),n.cur=a,"string.special"}function a(r,n){return r.match(/^HTTP\/\d\.\d$/)?(n.cur=p,"keyword"):e(r,n)}function p(r){return r.sol()&&!r.eat(/[ \t]/)?r.match(/^.*?:/)?"atom":(r.skipToEnd(),"error"):(r.skipToEnd(),"string")}function s(r){return r.skipToEnd(),null}t.r(n),t.d(n,{http:()=>d});const d={name:"http",token:function(r,n){var t=n.cur;return t!=p&&t!=s&&r.eatSpace()?null:t(r,n)},blankLine:function(r){r.cur=s},startState:function(){return{cur:u}}}}}]);
//# sourceMappingURL=3405.nbdime.js.map