"use strict";(self.webpackChunknbdime_webapp=self.webpackChunknbdime_webapp||[]).push([[272],{60272:(e,r,n)=>{n.r(r),n.d(r,{diff:()=>t});var s={"+":"inserted","-":"deleted","@":"meta"};const t={name:"diff",token:function(e){var r=e.string.search(/[\t ]+?$/);if(!e.sol()||0===r)return e.skipToEnd(),("error "+(s[e.string.charAt(0)]||"")).replace(/ $/,"");var n=s[e.peek()]||e.skipToEnd();return-1===r?e.skipToEnd():e.pos=r,n}}}}]);
//# sourceMappingURL=272.nbdime.js.map