

set -ex



pip check
nbdime show -h
nbshow -h
nbdiff -h
nbdiff-web -h
nbmerge -h
nbmerge-web -h
git-nbdifftool -h
git-nbmergetool -h
git-nbdiffdriver -h
git-nbmergedriver -h
hg-nbdiff -h
hg-nbdiffweb -h
hg-nbmerge -h
hg-nbmergeweb -h
jupyter labextension list
jupyter server extension list
jupyter labextension list 1>labextensions 2>&1
jupyter server extension list 1>server_extensions 2>&1
grep -iE "nbdime-jupyterlab.*OK*" labextensions
grep -iE "nbdime.*OK" server_extensions
exit 0
