# coding: utf-8

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

diff_ignorables = ('sources', 'outputs', 'attachments', 'metadata', 'id', 'details')
