{% extends "nbdimepage.html" %}

    {% block nbdimeheader %}

    <div id="nbdime-header" class="nbdime-Merge">
      <h3>Notebook Merge</h3>
      Enter notebook filenames or URLs in the form below to get started.

      <form id="nbdime-merge-form" class="nbdime-forms">
        <fieldset>
          <legend>Please input filenames/URLs of notebooks to merge:</legend>
          <label>Local:
          <input id="merge-local" type="text" name="local" value="{{ config_data['local']|e }}" /></label>
          <label>Base:
          <input id="merge-base" type="text" name="base" value="{{ config_data['base']|e }}" /></label>
          <label>Remote:
          <input id="merge-remote" type="text" name="remote" value="{{ config_data['remote']|e }}" /></label>
          <input id="nbdime-run-merge" type="submit" name="merge" value="Merge files" />
          <button id="nbdime-download" style="display: none" type="button">Download</button>
        </fieldset>
      </form> <!-- nbdime-forms -->
      <div id="nbdime-header-buttonrow">
        <label><input id="nbdime-hide-unchanged" type="checkbox"></input>Hide unchanged cells</label>
        <button id="nbdime-save" style="display: none">Save</button>
        <button id="nbdime-download" style="display: none">Download</button>
        <button id="nbdime-close" style="display: none">Close tool</button>
      </div>
      <div id=nbdime-header-banner>
        <span id="nbdime-header-local">Local</span>
        <span id="nbdime-header-base">Base</span>
        <span id="nbdime-header-remote">Remote</span>
      </div>
    </div> <!-- ndime-header -->

    {% endblock %}
