<!DOCTYPE html>
<html>

  <head>
    <meta charset="utf-8"/>
    <meta name="viewport" content="width=device-width, initial-scale=1.0" />

    <title>nbdime - diff and merge your Jupyter notebooks</title>

  </head>


  <!-- TODO: make nbdime.init() setup the forms/input user interface? -->

  <body> <!-- onload="nbdime.init()" -->

    {% block nbdimeheader %}
    {% endblock %}

    <div id="nbdime-root" class="nbdime-root">
    </div>

    <script id='nbdime-config-data' type="application/json">{{ config_data|tojson|safe }}</script>
    <script src="{{ static_url('nbdime.js') }}"></script>
    <noscript>Nbdime relies on javascript for diff/merge views!</noscript>
  </body>
</html>
