# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import operator

__all__ = ["diff_sequence_myers"]


def diff_sequence_myers(A, B, compare=operator.__eq__):
    """Compute the diff of A and B using Myers' O(ND) algorithm."""
    raise NotImplementedError("Work in progress...")
